package it.unitn.disi.ads.core.application;

import br.com.engenhodesoftware.util.TextUtils;
import it.unitn.disi.ads.core.application.exceptions.LoginFailedException;
import it.unitn.disi.ads.core.application.exceptions.LoginFailedReason;
import it.unitn.disi.ads.core.domain.Employee;
import it.unitn.disi.ads.core.persistence.EmployeeDAO;
import java.io.Serializable;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.faces.context.FacesContext;
import javax.inject.Inject;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;

/**
 * TODO: documentation pending.
 *
 * @author Vítor E. Silva Souza (vitorsouza@gmail.com)
 */
@Stateless
public class LoginService implements LoginServiceLocal, Serializable {
	private static final long serialVersionUID = 1L;

	/** The logger. */
	private static Logger logger = Logger.getLogger(LoginService.class.getCanonicalName());

	/** The DAO for Employee objects. */
	@EJB
	private EmployeeDAO employeeDAO;

	/** The current session's information. */
	@EJB
	private SessionInformation sessionInformation;

	/** @see it.unitn.disi.ads.application.LoginServiceLocal#login(java.lang.String, java.lang.String) */
	@Override
	public void login(String username, String password) throws LoginFailedException {
		try {
			// Retrieves the employee that owns the given username.
			Employee employee = employeeDAO.retrieveByUsername(username);

			// Creates the MD5 hash of the password for comparison.
			String md5pwd = TextUtils.produceMd5Hash(password);

			// Checks if the passwords match.
			String pwd = employee.getPassword();
			if ((pwd != null) && (pwd.equals(md5pwd))) {
				// Checks also if the container authenticates the user. If not, it will thrown an exception.
				HttpServletRequest request = (HttpServletRequest)FacesContext.getCurrentInstance().getExternalContext().getRequest();
				request.login(username, password);

				// Login successful. Registers the current user in the session.
				logger.log(Level.INFO, "Employee \"{0}\" successfully logged in.", username);
				sessionInformation.setCurrentUser(employee);
				pwd = null;
			}
			else {
				// Passwords don't match.
				logger.log(Level.INFO, "Employee \"{0}\" not logged in: password didn't match.", username);
				throw new LoginFailedException(LoginFailedReason.INCORRECT_PASSWORD);
			}
		}
		catch (NullPointerException e) {
			logger.log(Level.SEVERE, "Logging in employee \"" + username + "\" triggered an exception during acquisition of the HTTP request.", e);
			throw new LoginFailedException(LoginFailedReason.MD5_ERROR);
		}
		catch (ClassCastException e) {
			logger.log(Level.SEVERE, "Logging in employee \"" + username + "\" triggered an exception during acquisition of the HTTP request.", e);
			throw new LoginFailedException(LoginFailedReason.MD5_ERROR);
		}
		catch (NoSuchAlgorithmException e) {
			logger.log(Level.SEVERE, "Logging in employee \"" + username + "\" triggered an exception during MD5 hash generation.", e);
			throw new LoginFailedException(LoginFailedReason.MD5_ERROR);
		}
		catch (NoResultException e) {
			// No employee was found with the given username.
			logger.log(Level.INFO, "Employee \"{0}\" not logged in: no registered employee found with given username.", username);
			throw new LoginFailedException(e, LoginFailedReason.UNKNOWN_USERNAME);
		}
		catch (NonUniqueResultException e) {
			// Multiple employees were found with the same username.
			logger.log(Level.WARNING, "Employee \"{0}\" not logged in: there are more than one registered employee with the given username.", username);
			throw new LoginFailedException(e, LoginFailedReason.MULTIPLE_USERS);
		}
		catch (ServletException e) {
			// User was not authenticated by the container.
			logger.log(Level.WARNING, "Employee \"" + username + "\" not logged in: container returned an exception.", e);
			throw new LoginFailedException(e, LoginFailedReason.CONTAINER_REJECTED);
		}
	}
}
